from unittest import TestCase


class GetCaCert(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.test_quorum.AddDeviceNetTest
    """


class ClientSetup(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.test_quorum.AddDeviceNetTest
    """


class SignCertificate(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.test_quorum.AddDeviceNetTest
    """


class ClientImportCertificateAndKey(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.test_quorum.AddDeviceNetTest
    """


class ClientDestroy(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.test_quorum.RemoveDeviceNetTest
    """
