@font-face {
    font-family: 'Overpass';
    src: url('Overpass-Regular.ttf') format('truetype');
    font-weight: normal;
    font-style: normal;
}

@font-face {
    font-family: 'Overpass';
    src: url('Overpass-Bold.ttf') format('truetype');
    font-weight: bold;
    font-style: normal;
}

