<tr id="configure_title_row" {{bind-attr style="Pcs.configure_page"}}>
  <td id="page_header" colspan=3>
    <table id="config_options">
      <tr>
        <td>CLUSTER PROPERTIES</td>
      </tr>
    </table>
  </td>
</tr>
<tr id="configure_header_row" {{bind-attr style="Pcs.configure_page"}}>
  <td colspan=3>
    <hr>
  </td>
</tr>
<tr id="configure_list_row" {{bind-attr style="Pcs.configure_page"}}>
  <td id="config" colspan=3>
    {{#if Pcs.settingsController.error}}
      Unable to get cluster properties.<br/>
    {{else}}
      {{input
          type="text"
          value=Pcs.settingsController.filter
          placeholder="Filter"
      }}
      {{#if Pcs.settingsController.show_advanced}}
        <button onclick="Pcs.settingsController.set('show_advanced', false);">Hide advanced settings</button>
      {{else}}
        <button onclick="Pcs.settingsController.set('show_advanced', true);">Show advanced settings</button>
      {{/if}}
      <form id="cluster_properties">
        <table>
          {{#each property in Pcs.settingsController.filtered}}
            {{cluster-property prop=property name=property.name}}
          {{else}}
          <tr><td>No cluster properties available.</td></tr>
          {{/each}}
        </table>
        {{#if Pcs.settingsController.filtered}}
          <button onclick="update_cluster_settings(); return false;">
            Apply Changes
          </button>
        {{/if}}
      {{/if}}
      <button onclick="show_loading_screen(); refresh_cluster_properties(); return false;">
        Refresh
      </button>
    </form>
  </td>
</tr>
<tr>
  <td colspan=3>
    <hr>
  </td>
</tr>
